#!/usr/bin/perl
# Perl Pop3 Server
# Copyright (C) 2006 Sergio Camarena
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Version: 1.0.1
 
use Socket;
use Fcntl ':flock'; # import LOCK_* constants

close STDERR;
eval
{

# your domain or hostname
$servername = "perlpop3";

# Linux uses 'mail', *BSDs use 'system'.
$default_group = 'mail';

# /var/mail or /var/spool/mail are commonly used
$default_spool = "/var/mail";

$useDatabase = 0;

# UID to use when using database
$default_uid = 1;

# see DOC file for rest of options
$use_popauth = 0;
$update_relaydomains = 0;
$timeout = 120;
$logfile = '/var/log/poplog';
$debug = 0;

$max_unknown_commands = 5;

$database_connect = "dbi:Pg:dbname=popdb";
$database_user = 'root';
$database_password = '';

# NO more configurations are needed past this point

local $SIG{ALRM} = sub { print "-ERR timeout\r\n"; &unlock; exit(0); };

$time = time;
$| = 1;

if ( $hersockaddr = getpeername(STDIN) ) {

	($port, $iaddr) = sockaddr_in($hersockaddr);
	$herhostname    = gethostbyaddr($iaddr, AF_INET);
	$herstraddr     = inet_ntoa($iaddr);
	if ( $herstraddr =~ /^(\d+\.\d+\.\d+)\.\d+/ ) {
		$class = $1;
	}

}

if ( $useDatabase ) {
	print "+OK <$$.$time\@$servername>\r\n" || &printerror;
}
else {
	print "+OK\r\n" || &printerror;
}
&log( "ppop initiated: $herstraddr ($herhostname)" );

$authenticated = 0;
while( ! $authenticated ) {
alarm $timeout;
$authline = <>;
&log( "debug: $authline" ) if $debug;
my $count = 0;
alarm 0;
	if ( $authline =~ /^APOP\s+(.*?)\s+(.*)\s*$/i ) {
	    my $rpass;
		$user_input = lc( $1 );
		$user_input =~ s/[\/\\\0|]//g;
		eval( 'use Digest::MD5 qw( md5_hex )' );
		$md5pass = lc( $2 );
		$md5pass =~ s/[\n\r]//g;

		( $count, $username, $rpass, $domain, $spool, $uid ) = &get_user_data( $user_input );
		
		if ( $count == 0 ) {
			print "-ERR Password supplied is incorrect.\r\n" || &printerror;
			&log( "User not found: $username" );
		}
		elsif ( $count > 1 ) {
			print "-ERR Ambiguous username supplied. Please add ':yourdomain' to your username!\r\n" || &printerror;
			&log( "Ambig Username: $username" );
		}
		else
		{
			$spool = $nspool if $nspool;
			$md5right = md5_hex( "<$$.$time\@$servername>$rpass" );
			undef $rpass;
			if ( $md5pass eq $md5right ) {
				$authenticated = 1;
				last;
			}
			else
			{
				print "-ERR Password supplied is incorrect.\r\n" || &printerror;
				&log( "Invalid password: $username invalid: $authline" );
			}
		}
	}
	elsif ( $authline =~ /^USER\s+(.*)$/i ) {
		$user_input = lc( $1 );
		$user_input =~ s/[\/\\\0|\n\r]//g;
		print "+OK Please give password.\r\n" || &printerror;
		$username_given = 1;
		next;
	}
	elsif ( $authline =~ /^PASS\s+(.*)$/i && $username_given ) {
		$pass = $1;
		$pass =~ s/[\n\r]//g;
		( $count, $username, $rpass, $domain, $spool, $uid ) = &get_user_data( $user_input, $pass );

		if ( $count == 0 ) {
			print "-ERR Password supplied is incorrect.\r\n" || &printerror;
			&log( "User not found: $username" );
		}
		elsif ( $count > 1 ) {
			print "-ERR duplicate username, please contact administrator.\r\n" || &printerror;
			&log( "Duplicate username: $username" );
		}
		else
		{
			if ( $rpass eq $pass ) {
				$authenticated = 1;
				last;
			}
			else
			{
				$crypted = crypt( $pass, $rpass );
				if ( $rpass eq $crypted ) {
					$authenticated = 1;
					last;
				}

				print "-ERR Password supplied is incorrect.\r\n" || &printerror;
				&log( "Invalid password: $username invalid: $username" );
			}
		}
	
	}
	elsif ( $authline =~ /^QUIT\s*$/i )
	{
		print "+OK Bye.\r\n" || &printerror;
		exit(0);
	}
	elsif ( $authline =~ /^CAPA\s*$/i )
	{
		print "+OK Capability list follows\r\n" || &printerror;
		print "TOP\r\n" || &printerror;
		print "USER\r\n" || &printerror;
#		print "EXPIRE NEVER\r\n"; # can be "EXPIRE 0 USER"
#		print "SASL NTLM\r\n"; # NTLM -> Outlook KERBEROS_V4 SKEY SCRAM-MD5
#		print "RESP-CODES\r\n" || &printerror;
#		print "LOGIN-DELAY 60 USER\r\n"; # not implemented
#		print "PIPELINING\r\n"; # should work but should be roughly tested
		print "UIDL\r\n" || &printerror;
		print ".\r\n" || &printerror;
	}
	else
	{
	$authline =~ s/[\r\n]//g;
	&log( "Unknown command: $authline" );
	if ( ++$unknowncommands > $max_unknown_commands ) {
		&log( "Maximum unknown commands reached" );
		print "-ERR error\r\n" || &printerror;
		exit;
		}
	print "-ERR Unknown command.\r\n" || &printerror;

	}
	
	$username_given = 0; # reset USER login
}

if ( ! $authenticated ) {
	print "-ERR Not authorized\r\n" || &printerror;
	&log( "NOT AUTHENTICATED!" );
	exit( 0 );
	}

$0 = "ppop '$username'";

if ( $use_popauth ) {
## add ip to popauth  database

  eval 'use DB_File';

  tie %popdb, "DB_File", "/etc/mail/popauth.db", O_RDWR, 0640, $DB_HASH || die $@;
  $lastip = $popdb{ $username };
  delete $popdb{ "$lastip" };
  $popdb{ "$herstraddr" } = "OK";
  $popdb{ $username } = $herstraddr;

}

if ( $update_relaydomains ) {
	open( REL, "/etc/mail/relay-domains" );
	    while ( <REL> ) {
			if ( /^$class/ ) {
				$relaydomain_found = 1;
				last;
			}
	    }
	close REL;
	
	if ( ! $relaydomain_found ) {
		open( REL, ">>/etc/mail/relay-domains" ) || &printerror;
		print REL "$class\n";
		close REL;
		system( "/usr/local/bin/zap -s 1 -y sendmail > /dev/null 2> /dev/null" );
		$extramessage = "(Class $class accepted)";
	}
}

&log( "$username checked mail" );

if ( open( LOCK, ">$spool/$username.lock" ) ) {
	print LOCK $$;
	flock (LOCK,LOCK_EX);
}
else
{
	print "-ERR Could not lock mailbox. Please check with system administrator.\r\n" || &printerror;
	&log( "Mailbox locked: $username Could not open lock" );
	exit(0);
}

&get_spool_info( $username );

print "+OK $numbermessages message(s) ($totalsize octets). $extramessage\r\n" || &printerror;
alarm $timeout;
while( $line = <> ) {
	&log( "debug: $line" ) if $debug;

	alarm 0;
	if ( $line =~ /^LIST\s*$/i ) {
		print "+OK $numbermessages message(s) ($totalsize octets).\r\n" || &printerror;
		@mailsizestmp = @mailsizes;
		while( $#mailsizestmp+1 ) {
			$number = shift @mailsizestmp;
			$size = shift @mailsizestmp;
			print "$number $size\r\n" unless $deleted{ $number };
			}
			print ".\r\n" || &printerror;

	}
	elsif ( $line =~ /^LIST\s+(\d+)\s*$/i ) {
		$messagetolist = $1;
		if ( $mailsizes[-1+2*$messagetolist] == 0 || $messagetolist < 1 || $deleted{ $messagetolist } ) {
			print "-ERR no such message or message deleted.\r\n" || &printerror;
		}
		else
		{
			print "+OK $messagetolist ".$mailsizes[-1+2*$messagetolist]."\r\n" || &printerror;
		}

	}
	elsif ( $line =~ /^UIDL\s*$/i ) {
		print "+OK UIDL list follows\r\n" || &printerror;
		foreach $msg ( sort { $a <=> $b } keys %uidl ) {
			print "$msg $uidl{ $msg }\r\n" unless $deleted{ $msg };
			}
		print ".\r\n" || &printerror;
	}
	elsif ( $line =~ /^UIDL\s+(\d+)\s*$/i ) {
		$messagetolist = $1;
		if ( $mailsizes[-1+2*$messagetolist] == 0 || $messagetolist < 1 || $deleted{ $messagetolist } ) {
			print "-ERR no such message or message deleted.\r\n" || &printerror;
		}
		else
		{
			print "+OK $messagetolist ".$uidl{$messagetolist}."\r\n" || &printerror;
		}

	}
	elsif ( $line =~ /^STAT\s*$/i ) {
		print "+OK $numbermessages $totalsize\r\n" || &printerror;
		&log( "debug: +OK $numbermessages $totalsize\r\n" ) if $debug;
		}
	elsif ( $line =~ /^QUIT\s*$/i ) {
	    my @deleted = keys %deleted;
	    if ( @deleted || $uidl_changed ) {
		&log( "debug: quit received, spool changed, resaving spool file\r\n" ) if $debug;
		&save_mail;
		}
	    print "+OK Bye.\r\n" || &printerror;
	    &log( "debug: quit received" ) if $debug;
	    &unlock;
	    exit(0);
	}
	elsif ( $line =~ /^RETR\s+(\d+)\s*$/i ) {
		my $messagetoret = $1;
		if ( $mailsizes[-1+2*$messagetoret] == 0 || $messagetoret < 1 || $deleted{ $messagetoret } ) {
			print "-ERR no such message or message deleted.\r\n" || &printerror;
		}
		else
		{
		print "+OK Message follows.\r\n" || &printerror;
		
		&log( "debug: RETR received for '$messagetoret'.\r\n" ) if $debug;
		&printmessage( $messagetoret );
		print "\r\n.\r\n" || &printerror;
		}
	}
	elsif ( $line =~ /^TOP\s+(\d+)\s+(\d+)\s*$/i ) {
		my $messagetoret = $1;
		$lines = $2;
		if ( $mailsizes[-1+2*$messagetoret] == 0 || $messagetoret < 1 || $deleted{ $messagetoret } ) {
			print "-ERR no such message or message deleted.\r\n" || &printerror;
		}
		elsif( $lines < 0 )
		{
			print "-ERR Lines must be positive.\n" || &printerror;
		}
		else
		{
		print "+OK Message follows.\r\n" || &printerror;
		&topmessage( $messagetoret, $lines );
		print "\r\n.\r\n" || &printerror;
		}
	}
	elsif ( $line =~ /^DELE\s+(\d+)\s*$/i ) {
		my $messagetodel = $1;
		if ( $mailsizes[-1+2*$messagetodel] == 0 || $messagetodel < 1 ) {
			print "-ERR No such message.\r\n" || &printerror;
		}
		elsif ( $deleted{ $messagetodel } )
		{
			print "-ERR Message already deleted.\r\n" || &printerror;
		}
		else
		{
		$deleted{ $messagetodel } = 1;
		$numbermessages--;
		$totalsize -= $mailsizes[-1+2*$messagetodel];
		print "+OK Message Deleted\r\n" || &printerror;
		}
	}
	elsif ( $line =~ /^NOOP\s*$/i ) {
		print "+OK\r\n" || &printerror;
	}
	elsif ( $line =~ /^rset\s*$/i ) {
		foreach $msg ( keys %deleted ) {
			$numbermessages++;
			$totalsize += $mailsizes[-1+2*$msg];
			}
		undef %deleted;
		print "+OK $numbermessages message(s) ($totalsize octets).\r\n" || &printerror;
	}
	elsif ( $line =~ /^CAPA\s*$/i ) {
		print "+OK Capability list follows\r\n" || &printerror;
		print "TOP\r\n" || &printerror;
		print "USER\r\n" || &printerror;
		print "EXPIRE NEVER\r\n" || &printerror; # Could be "EXPIRE 0 USER" (does not permit but varies by user)
#		print "SASL NTLM\r\n"; # NTLM -> Outlook KERBEROS_V4 SKEY SCRAM-MD5
#		print "RESP-CODES\r\n" || &printerror;
#		print "LOGIN-DELAY 10\r\n"; # not implemented until enforced / varies with user
#		print "PIPELINING\r\n"; #should be toughly beta-tested
		print "UIDL\r\n" || &printerror;
		print "IMPLEMENTATION SACM/1.0\r\n" || &printerror;
		print ".\r\n" || &printerror;

	}
	else {
		$line =~ s/[\r\n]//g;
		print "-ERR Unknown command or not implemented\r\n" || &printerror;
		&log( "Unknown command: $line" );
	}
	alarm $timeout;
}

};

sub get_spool_info
{
$numbermessages = 0;
$totalsize = 0;
$prevlineempty = 1;
if ( ! -e "$spool/$username" ) {
  if ( ! open( MAILFILE, ">$spool/$username" ) ) {
    print( "-ERR Could not create spool file!" ) || &printerror;
    &log( "Could not open spool file: '$spool/$username'!" );
    &unlock;
    exit(1);
  }
  close MAILFILE;
}

if ( ! open( MAILFILE, "$spool/$username" ) ) {
  print( "-ERR Could not open spool file: '$spool/$username!" ) || &printerror;
  &log( "Could not open spool file: '$spool/$username'!" );
  &unlock;
  exit(1);
}
$veryfirstline = 1;
while( <MAILFILE> ) {
	s/\n/\r\n/;
	s/^\./\.\./;
	if ( ( /^From /  && $prevlineempty ) || $veryfirstline ) {
    	    $veryfirstline = 0;
		if ( $numbermessages ) {
		   if ( ! $uidl{ $numbermessages } ) {
			$uidlchanged{ $numbermessages } = 1;
			$uidlchanged = 1;

			do {
			$uidl{ $numbermessages } = time.int( rand() * 1E4 );
			} until $useduidl{ $uidl{ $numbermessages } } == 0;

			$useduidl{ $uidl{ $numbermessages } } = 1;
		   }

		push @mailsizes, $numbermessages;
		push @mailsizes, $totalsize-$prevsize;
		}
		$prevsize = $totalsize;
		$numbermessages++;
	}
	elsif ( /X-UIDL: (\d+)/ ) {
		$uidl{ $numbermessages } = $1;
		$useduidl{ $uidl{ $numbermessages } } =	1;
	}

	$totalsize += length( $_ );
	/^\r?$/ ? { $prevlineempty = 1 } : { $prevlineempty = 0 };
	}
if ( $numbermessages ) {
			if ( ! $uidl{ $numbermessages } ) {
			$uidlchanged{ $numbermessages } = 1;
			$uidl_changed = 1;
			do {
			$uidl{ $numbermessages } = time.int( rand() * 1E4 );
			} until $useduidl{ $uidl{ $numbermessages } } == 0;

			$useduidl{ $uidl{ $numbermessages } } = 1;
			}
push @mailsizes, $numbermessages;
push @mailsizes, $totalsize-$prevsize;
}
if ( ! close MAILFILE ) {
  print( "-ERR Error close spoolfile\r\n" ) || &printerror;
  &log( "Could not close spoolfile '$spool/$username'!" );
  &unlock;
  exit( 1 );
}

}

sub printmessage
{
my $noopennec;
my $nm = 0;
my $doprint = 0;
$prevlineempty = 1;
my ( $messagetoret ) = @_;
if ( $previously != 0 && $messagetoret > $previously ) {
	$noopennec = 1;
	$nm = $previously;
	$nm++;
	if ( $nm == $messagetoret ) {
		print $lastline || &printerror;
		$doprint = 1;
		}
	$lastline =~ /^\r?$/ ? { $prevlineempty = 1 } : { $prevlineempty = 0 };
}

if ( ! $noopennec ) {
$veryfirstline = 1;
if ( ! open( MAILFILE, "$spool/$username" ) ) {
  print( "-ERR Could not open spool file: '$spool/$username!" ) || &printerror;
  &log( "Could not open spool file: '$spool/$username'!" );
  exit(1);
}
}
$| = 1024;
while( <MAILFILE> ) {
	s/\n/\r\n/;
	s/^\./\.\./;
	if ( ( /^From / && $prevlineempty ) || $veryfirstline ) {
                $veryfirstline = 0;
		last if $doprint;
		$nm++;
		if ( $nm == $messagetoret ) {
			print || &printerror;
			$doprint = 1;
			}
		}
		else
		{
		if ( $doprint ) {
			print || &printerror;
			}
		}
	/^\r?$/ ? { $prevlineempty = 1 } : { $prevlineempty = 0 };

	}
$| =1;
$previously = $messagetoret;
$lastline = $_;
}

sub topmessage
{
my ( $messagetoret, $lines ) = @_;
my $nm = 0;
my $doprint = 0;
my $printed = -1;
my $headers_printed = 0;
$prevlineempty = 1;
if ( ! open( MAILFILE, "$spool/$username" ) ) {
  print( "-ERR Could not open spool file: '$spool/$username!" ) || &printerror;
  &log( "Could not open spool file: '$spool/$username'!" );
  exit(1);
}
$veryfirstline = 1;
while( <MAILFILE> ) {
	s/\n/\r\n/;
	s/^\./\.\./;
	if ( ( /^From / && $prevlineempty ) || $veryfirstline ) {
	        $veryfirstline = 0;
		last if $doprint;
		$nm++;
		if ( $nm == $messagetoret ) {
			print || &printerror;
			$doprint = 1;
			}
		}
		else
		{
		if ( $doprint ) {
			print || &printerror;
		}
		if ( /^\r?$/ && $doprint ) { $headers_printed = 1 }
		if ( $headers_printed && $doprint ) {
			if ( ++$printed == $lines )
			{
			last;
			}
		    }
		}
	/^\r?$/ ? { $prevlineempty = 1 } : { $prevlineempty = 0 };
	}
close MAILFILE;

}

sub save_mail
{
		&log( "debug: saving...\r\n" ) if $debug;
		if ( ! open( FIN, ">$spool/$username.tmp" ) ) {
		        &unlock;
			print "-ERR Error creating temp file.\r\n" || &printerror;
			exit(0);
			}

                if ( ! open( MAILFILE, "$spool/$username" ) ) {
                  print ( "-ERR Could not open spool file: '$spool/$username!" ) || &printerror;
                  &log( "Could not open spool file: '$spool/$username'!" );
                  exit(1);
                }
                
		$nm = 0;
		$prevlineempty = 1;
		$veryfirstline = 1;
		while( <MAILFILE> ) {
		if ( ( /^From / && $prevlineempty ) || $veryfirstline ) {
		        $veryfirstline = 0;
			$nm++;
			$headerprinted = 0; $uidlfound = 0;
			print FIN if ! $deleted{ $nm };
		}
		elsif ( /^X-UIDL: / ) {
		        $uidlfound = 1;
		        print FIN if ! $deleted{ $nm };
		
		}
		elsif ( ! $deleted{ $nm } ) {
			if ( /^$/ && $uidlchanged{ $nm } && ! $headerprinted && ! $uidlfound ) {
				print FIN "X-UIDL: $uidl{ $nm }\n";
				$headerprinted = 1;
				}
			print FIN;
			}
		/^$/ ? { $prevlineempty = 1 } : { $prevlineempty = 0 };
		}
		close MAILFILE;


		if ( ! unlink( "$spool/$username" ) ) {
			print "-ERR Error updating spool.\r\n" || &printerror;
			&unlock;
			exit(0);
			}


		$system_gid = (getpwnam( $default_group ))[3];
		chown $uid,$system_gid, "$spool/$username.tmp";

		chmod 0600, "$spool/$username.tmp";
		if ( ! rename( "$spool/$username.tmp", "$spool/$username" ) ) {
			print "-ERR Could not update spool file!\r\n" || &printerror;
			&unlock;
			exit(0);
			}

		if ( ! close( FIN ) ) {
			unlink( "$spool/$username.lock" );
			print "-ERR Error writing to temp file.\r\n" || &printerror;
			&unlock;
			exit(0);
			}

}


sub db_connect
{
eval( 'use DBI' );
$db = DBI->connect($database_connect, $database_user, $database_password, { RaiseError => 1, AutoCommit => 1 }) || die $DBI::errstr;
}


sub db_disconnect
{
$db->disconnect;
}

sub log
{
my ( $msg ) = @_;
($sec, $min, $hour, $mday, $mon, $year ) = localtime( time );
$year += 1900; $mon++;
$strdate = sprintf( "%d/%.2d/%.2d %.2d:%.2d:%.2d", $year, $mon, $mday, $hour, $min, $sec );

if ( ! open( LOG, ">>$logfile" ) ) {
  print "-ERR Error opening log file: $logfile\r\n";
  &unlock;
  exit(1);
}

$msg =~ s/[\n\r]//g;
print LOG "$strdate [$$] $msg\r\n";
if ( ! close LOG ) {
  print "-ERR Error closing log file: $logfile!\r\n";
  }
}

sub printerror
{
&log( "Print out error. Exiting\r\n" );
&unlock;
exit(1);

}

sub unlock
{
  close LOCK;
  unlink "$spool/$username.lock";
}

	
sub get_user_data {
	my ( $count, $username, $rpass, $domain, $spool );
	my ( $user_input, $pass ) = @_;

	$user_input =~ s/'/''/g;
	$user_input =~ s/\\/\\\\/g;
	
	if ( $user_input =~ /(.*)[:\@](.*)/ ) {
		$username = $1;
		$domain = $2;
	}
			
	if ( $useDatabase ) {
		&db_connect;

		$pass =~ s/'/''/g;
		$pass =~ s/\\/\\\\/g;
		$domain =~ s/'/''/g;
		$domain =~ s/\\/\\\\/g;

		if ( $pass ) {
			$sql = "SELECT password, spool FROM users WHERE username = '$username' AND password = '$pass'";
		}
		elsif ( $domain ) {
			$sql = "SELECT password, spool FROM users WHERE username = '$username' AND domain = '$domain'";
		}
		else {
			$sql = "SELECT password, spool FROM users WHERE username = '$user_input'";			
		}
		$cur = $db->prepare( $sql );
		$cur->execute();
		while ( @fields = $cur->fetchrow ) {
	        ( $rpass, $spool ) = @fields;
			$count++;
		}
		$cur->finish;
		
		&db_disconnect;
		$uid = $default_uid;
	}
	else {
		$count = 1;
		$spool = $default_spool;
		( $username, $rpass, $uid ) = getpwnam( $user_input );
	}
	
	return( $count, $username, $rpass, $domain, $spool, $uid );
}
